import re
import socket
import threading

import pika

from clitellum.core import compressors, loggerManager
from clitellum.endpoints.channels import reconnectiontimers

__author__ = 'sergio'

class AmqpHostConfiguration(object):
    '''
    Configuracion del host de amqp
    '''
    def __init__(self):
        self._server = None
        self._port = 5672
        self._exchange = ExchangeConfiguration()
        self._user = 'guest'
        self._password = 'guest'
        self._queue = QueueConfiguration()
        self._use_ack = True
        self._qos = QosConfiguration()

    def get_server(self):
        '''
        Devuelve Nombre del host o direccion ip del servidor amqp
        '''
        return self._server

    def set_server(self, value):
        '''
        Establece el host o direccion ip del servidor amqp
        '''
        self._server = value

    def get_port(self):
        '''
        Numero del puerto de conexion del servidor amqp, por defecto 5672
        '''
        return self._port

    def set_port(self, value):
        '''
        Establece el puerto de conexion con el servidor amqp
        '''
        self._port = value

    def get_exchange(self):
        '''
        Exchange de conexion del servidor amqp
        '''
        return self._exchange

    def get_user(self):
        '''
        Devuelve nombre del usuario para conectar al servidor amqp
        '''
        return self._user

    def set_user(self, value):
        '''
        Establece nombre del usuario para conectar al servidor amqp
        '''
        self._user = value

    def get_password(self):
        '''
        Contraseña del usuario para conectar al servidor amqp
        '''
        return self._password

    def set_password(self, value):
        '''
        Establece la contraseña del usuario para conectar al servidor amqp
        '''
        self._password = value

    def get_queue(self):
        '''
        Devuelve la configuracion de la queue
        '''
        return self._queue

    def set_use_ack(self, value):
        self._use_ack = value

    def get_use_ack(self):
        return self._use_ack

    def get_qos(self):
        return self._qos
    
class ExchangeConfiguration(object):
    '''
    Configuracion del exchange de amqp
    '''
    def __init__(self):
        self._name = None
        self._type = ExchangeType.TOPIC
        self._create = True

    def get_name(self):
        '''
        Nombre del exchange
        '''
        return self._name

    def set_name(self, value):
        '''
        Establece el nombre del exchange
        '''
        self._name = value

    def get_type(self):
        '''
        Tipo de exchange
        '''
        return self._type

    def set_type(self, value):
        '''
        Tipo de exchange
        '''
        self._type = value

    def must_be_created(self):
        '''
        Indica si el exchange debe crearse en el momento de la conexion, por defecto True
        '''
        return self._create

    def set_created(self, value):
        '''
        Indica si el exchange debe crearse en el momento de la conexion, por defecto True
        '''
        self._create = value

class QueueConfiguration(object):
    '''
    Configuracion de la cola de amqp
    '''
    def __init__(self):
        self._name = None
        self._create = True
        self._routing_keys = []

    def get_name(self):
        '''
        Nombre de la cola donde conectarse
        '''
        return self._name

    def set_name(self, value):
        '''
        Establece el nombre de la cola donde conectarse
        '''
        self._name = value

    def get_routing_keys(self):
        '''
        Devuelve la lista de routing keys
        '''
        return self._routing_keys

    def add_routing_key(self, key):
        '''
        Inserta una nueva clave de enrutamiento para la cola
        '''
        self._routing_keys.append(key)

    def add_routing_keys(self, keys):
        '''
        Inserta un grupo de claves de enrutamiento
        '''
        for key in keys:
            self.add_routing_key(key)

    def must_be_created(self):
        '''
        Indica si el exchange debe crearse en el momento de la conexion, por defecto True
        '''
        return self._create

    def set_created(self, value):
        '''
        Indica si el exchange debe crearse en el momento de la conexion, por defecto True
        '''
        self._create = value

class QosConfiguration(object):
    '''
    Configuracion de la calidad del servicio
    '''

    def __init__(self):
        self._prefetch_size = 0
        self._prefetch_count = 10000

    def set_prefetch_size(self, size):
        '''
        Establece el prefetch size del amqp
        '''
        self._prefetch_size = size

    def get_prefetch_size(self):
        '''
        Devuelve el prefetch size del amqp
        '''
        return self._prefetch_size

    def set_prefetch_count(self, count):
        '''
        Establece el prefetch count del amqp
        '''
        self._prefetch_count = count

    def get_prefetch_count(self):
        '''
        Devuelve el prefetch count del amqp
        '''
        return self._prefetch_count

class ExchangeType(object):
    '''
    Enumeracion con los tipos posibles de exchange
    '''
    DIRECT = 'direct'
    TOPIC = 'topic'

class AmqpHostConfigurationBuilder(object):
    '''
    Clase que genera un constructor de la configuracion del host
    '''

    OUT_BOUND_CHANNEL = 'outboundchannel'
    IN_BOUND_CHANNEL = 'inboundchannel'

    @classmethod
    def create(cls):
        return AmqpHostConfigurationBuilder()

    def __init__(self):
        self._uri = None
        self._exchange_type = 'direct'
        self._type = AmqpHostConfigurationBuilder.OUT_BOUND_CHANNEL
        self._create_exchange = True
        self._create_queue = True
        self._user = 'guest'
        self._password = 'guest'
        self._use_ack = True
        self._prefetch_size = 0
        self._prefetch_count = 10000

    def set_uri(self, uri):
        '''
        Establece la uri del servidor de amqp
        '''
        self._uri = uri
        return self

    def set_direct_exchange(self):
        '''
        Establece el tipo direct en el exchange
        '''
        self._exchange_type = ExchangeType.DIRECT
        return self

    def set_topic_exchange(self):
        '''
        Establece el tipo topic en el exchange
        '''
        self._exchange_type = ExchangeType.TOPIC
        return self

    def set_create_exchange(self, value):
        '''
        Establece que se cree el exchange durante la conexion
        '''
        self._create_exchange = value
        return self

    def set_create_queue(self, value):
        '''
        Establece que se cree el cola durante la conexion
        '''
        self._create_queue = value
        return self

    def set_is_out_bound(self):
        '''
        Indica que la uri es para un canal de salida
        '''
        self._type = AmqpHostConfigurationBuilder.OUT_BOUND_CHANNEL
        return self

    def set_is_in_bound(self):
        '''
        Indica que la uri es para un canal de entrada
        '''
        self._type = AmqpHostConfigurationBuilder.IN_BOUND_CHANNEL
        return self

    def set_credentials(self, user, password):
        '''
        Establece la credenciales de conexion
        '''
        self._user = user
        self._password = password
        return self

    def set_use_ack(self, value):
        self._use_ack = value
        return self

    def set_prefetch_count(self, count):
        '''
        Establece el prefetch count del amqp
        '''
        self._prefetch_count = count
        return self
    
    def set_prefetch_size(self, size):
        '''
        Establece el prefetch size del amqp
        '''
        self._prefetch_size = size
        return self

    def build(self):
        '''
        Crea una instancia de la configuracion del servidor
        '''
        ret = AmqpHostConfiguration()
        ret.set_user(self._user)
        ret.set_password(self._password)
        ret.set_use_ack(self._use_ack)

        if self._type == AmqpHostConfigurationBuilder.OUT_BOUND_CHANNEL:
            self._extract_exchange(ret)
        else:
            if self._exchange_type == ExchangeType.DIRECT:
                self._extract_queue_direct(ret)
            else:
                self._extract_queue_topic(ret)

        ret.get_exchange().set_type(self._exchange_type)
        ret.get_exchange().set_created(self._create_exchange)
        ret.get_queue().set_created(self._create_queue)
        ret.get_qos().set_prefetch_count(self._prefetch_count)
        ret.get_qos().set_prefetch_size(self._prefetch_size)
        return ret

    def _extract_exchange(self, host_config: AmqpHostConfiguration):
        match = re.search(r'^amqp://(.*):(\d+)/(.*)', self._uri)
        if match:
            host_config.set_server(match.group(1))
            host_config.set_port(int(match.group(2)))
            host_config.get_exchange().set_name(match.group(3))
        else:
            raise NameError("Invalid host name", self._uri)

    def _extract_queue_topic(self, host_config: AmqpHostConfiguration):

        match = re.search(r'^amqp://(.*):(\d+)/(.*)/(.*)/(.*)', self._uri)
        if match:
            host_config.set_server(match.group(1))
            host_config.set_port(int(match.group(2)))
            host_config.get_exchange().set_name(match.group(3))
            host_config.get_queue().set_name(match.group(4))
            host_config.get_queue().add_routing_keys(match.group(5).split(';'))
        else:
            raise NameError("Invalid host name", self._uri)

    def _extract_queue_direct(self, host_config: AmqpHostConfiguration):

        match = re.search(r'^amqp://(.*):(\d+)/(.*)/(.*)', self._uri)
        if match:
            host_config.set_server(match.group(1))
            host_config.set_port(int(match.group(2)))
            host_config.get_exchange().set_name(match.group(3))
            host_config.get_queue().set_name(match.group(4))
            host_config.get_queue().add_routing_key(match.group(4))
        else:
            raise NameError("Invalid host name", self._uri)




from clitellum.core import compressors
from clitellum.endpoints.channels import reconnectiontimers
from clitellum.endpoints.channels.basechannels import Channel
from clitellum.endpoints.channels.amqp.channels import InBoundAmqpChannel, OutBoundAmqpChannel
from clitellum.endpoints.channels.amqp.configuration import AmqpHostConfigurationBuilder

class AmqpChannelBuilderFactory(object):
    '''
    Factoria de builders
    '''
    @classmethod
    def create_in_bound_builder(cls):
        '''
        Devuelve una instancia del builder de canales de entrada amqp
        '''
        reconnection_timer = reconnectiontimers.CreateLogarithmicTimer()
        amqp_host_builder = AmqpHostConfigurationBuilder.create()
        compressor = compressors.NoCompressor()

        return InBoundAmqpChannelBuilder(reconnection_timer, amqp_host_builder, compressor)

    @classmethod
    def create_out_bound_builder(cls):
        '''
        Devuelve una instancia del builder de canales de salida amqp
        '''
        reconnection_timer = reconnectiontimers.CreateLogarithmicTimer()
        amqp_host_builder = AmqpHostConfigurationBuilder.create()
        compressor = compressors.NoCompressor()

        return OutBoundAmqpChannelBuilder(reconnection_timer, amqp_host_builder, compressor)
    

class AmqpChannelBuilder():
    '''
    Constructor de canales Amqp
    '''
    def __init__(self, reconnection_timer, amqp_host_builder, compressor):
        self._reconnection_timer = reconnection_timer
        self._amqp_host_builder = amqp_host_builder
        self._compressor = compressor
        self._max_reconnections = Channel.MAX_RECONNECTIONS


    def set_uri(self, uri):
        '''
        Establece la uri del servidor de amqp
        '''
        self._amqp_host_builder.set_uri(uri)
        return self

    def set_direct_exchange(self):
        '''
        Establece el tipo direct en el exchange
        '''
        self._amqp_host_builder.set_direct_exchange()
        return self

    def set_max_recconections(self, max_reconnections):
        '''
        Establece el numero maximas de reconexiones contra el servidor de amqp
        por defecto son infinitas
        '''
        self._max_reconnections = max_reconnections
        return self

    def set_topic_exchange(self):
        '''
        Establece el tipo topic en el exchange
        '''
        self._amqp_host_builder.set_topic_exchange()
        return self

    def set_create_exchange(self, value):
        '''
        Establece que se cree el exchange durante la conexion
        '''
        self._amqp_host_builder.set_create_exchange(value)
        return self

    def set_create_queue(self, value):
        '''
        Establece que se cree el cola durante la conexion
        '''
        self._amqp_host_builder.set_create_queue(value)
        return self

    def set_credentials(self, user, password):
        '''
        Establece la credenciales de conexion
        '''
        self._amqp_host_builder.set_credentials(user, password)
        return self

    def set_use_ack(self, value):
        '''
        Establece si se debe usar ack en la conexion
        '''
        self._amqp_host_builder.set_use_ack(value)
        return self

    def set_prefetch_count(self, count):
        '''
        Establece el prefetch count del amqp
        '''
        self._amqp_host_builder.set_prefetch_count(count)
        return
    
    def set_prefetch_size(self, size):
        '''
        Establece el prefetch size del amqp
        '''
        self._amqp_host_builder.set_prefetch_size(size)
        return self

    def set_reconnection_timer(self, reconnection_timer):
        '''
        Establece el temporizador de la reconexion
        por defecto Logaritmico
        '''
        self._reconnection_timer = reconnection_timer
        return self

    def set_compressor(self, compressor):
        '''
        Establece el compresor del endpoint
        '''
        self._compressor = compressor
        return self

    def build(self):
        '''
        Crear una instancia de un canal Amqp
        '''
        raise NotImplementedError()

class OutBoundAmqpChannelBuilder(AmqpChannelBuilder):
    '''
    Constructor de canales de salida ampp
    '''
    def __init__(self, reconnection_timer, amqp_host_builder, compressor):
        AmqpChannelBuilder.__init__(self, reconnection_timer, amqp_host_builder, compressor)

    def build(self):
        amqp_host_conf = self._amqp_host_builder.set_is_out_bound().build()
        channel = OutBoundAmqpChannel(amqp_host_conf, self._reconnection_timer, \
                                    self._max_reconnections, \
                                    self._compressor)
        return channel


class InBoundAmqpChannelBuilder(AmqpChannelBuilder):
    '''
    Constructor de canales de entrada amqp
    '''

    def __init__(self, reconnection_timer, amqp_host_builder, compressor):
        AmqpChannelBuilder.__init__(self, reconnection_timer, amqp_host_builder, compressor)
        self._recception_timeout = 10
        self._max_threads = 1

    def set_timeout(self, timeout):
        '''
        Establece el timeout del bucle de recepcion, por defecto son 10 segundos
        '''
        self._recception_timeout = timeout
        return self

    def set_max_threads(self, max_threads):
        '''
        Establece el numero de hilos de extraccion
        '''
        self._max_threads = max_threads
        return self


    def build(self):
        '''
        Crear una instancia de un canal Amqp
        '''
        amqp_host_conf = self._amqp_host_builder.set_is_in_bound().build()
        channel = InBoundAmqpChannel(amqp_host_conf, self._reconnection_timer, \
                                    self._max_reconnections, self._recception_timeout, \
                                    self._compressor, self._max_threads)
        return channel



# author = sbermudel
# package description

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
import gzip
import zlib
import bz2

DEFAULT_ENCODING='utf-8'

def DefaultCompressor():
    return NoCompressor()


## Metodo factory que crea un compresor
# @param type Tipo de compresion [Gzip, Zip, Bz2, None]
# @param compressLevel Nivel de compresion, valor comprendido entre 0 y 9
def CreateCompressor(type_compressor, compressLevel=9, encoding=DEFAULT_ENCODING):

    compressor = None

    if type_compressor.lower() == "gzip":
        compressor = GzipCompressor(compressLevel)
    elif type_compressor.lower() == "zip":
        compressor = ZipCompressor(compressLevel)
    elif type_compressor.lower() == "bz2":
        compressor = Bz2Compressor(compressLevel)
    elif type_compressor.lower() == "none":
        compressor = NoCompressor()
    else:
        compressor = DefaultCompressor()
    compressor.set_encoding(encoding)
    return compressor


def CreateCompressorFromConfig(config):
    '''
    Crea un compresor a partir de un diccionario
    '''
    level = 6
    if 'compressLevel' in config:
        level = int(config['compressLevel'])

    encoding = DEFAULT_ENCODING
    if 'encoding' in config:
        encoding = config['encoding']

    return CreateCompressor(config["type"], level, encoding)

## Clase que define los metodos basicos de un compresor en memoria
class Compressor:
    ## Crea una instacia de Compressor
    # @param compressLevel Nivel de compresion
    def __init__(self, compressLevel):
        self.__compressLevel = compressLevel
        self._encoding = DEFAULT_ENCODING

    ## Devuelve el nive de compresion del compressor
    @property
    def compressLevel(self):
        return self.__compressLevel

    def set_encoding(self, encondig):
        '''
        Establece el encoding del compresor
        '''
        self._encoding = encondig

    def get_encoding(self):
        '''
        Devuelve el encoding del compresor
        '''
        return self._encoding


    ## Comprime el buffer especificado
    def compress(self, buffer):
        return buffer

    ## Descomprime el buffer especificado
    def decompress(self, buffer):
        return buffer


## Implementacion de un no compresor
class NoCompressor(Compressor):
    ## Crea una instacia de NoCompressor
    def __init__(self):
        Compressor.__init__(self, None)

    ## Comprime el buffer especificado
    def compress(self, buffer):
        return bytes(buffer, self.get_encoding())

    ## Descomprime el buffer especificado
    def decompress(self, buffer):
        return str(buffer, self.get_encoding())

## Implemtacion de la compresion Gzip
class GzipCompressor(Compressor):
    ## Crea una instacia de GzipCompressor
    # @param compressLevel Nivel de compresion, debe ser un numero entero entre 1 y 9, por defecto 9
    def __init__(self, compressLevel=9):
        Compressor.__init__(self, compressLevel)

    def compress(self, buffer):
        return gzip.compress(bytes(buffer, self.get_encoding()))

    def decompress(self, buffer):
        return str(gzip.decompress(buffer), self.get_encoding())


## Implemtacion de la compresion Gzip
class ZipCompressor(Compressor):
    ## Crea una instacia de ZipCompressor
    # @param compressLevel Nivel de compresion, debe ser un numero entero entre 0 y 9, por defecto 6
    def __init__(self, compressLevel=6):
        Compressor.__init__(self, compressLevel)

    def compress(self, buffer):
        return zlib.compress(bytes(buffer, self.get_encoding()), level=self.compressLevel)

    def decompress(self, buffer):
        return str(zlib.decompress(buffer), self.get_encoding())


## Implementacion de la compresion en BZ2
class Bz2Compressor(Compressor):
    ## Crea una instacia de Bz2Compressor
    # @param compressLevel Nivel de compresion, debe ser un numero entero entre 1 y 9, por defecto 9
    def __init__(self, compressLevel=9):
        Compressor.__init__(self, compressLevel=compressLevel)

    def compress(self, buffer):
        return bz2.compress(bytes(buffer, self.get_encoding()), compresslevel=self.compressLevel)

    def decompress(self, buffer):
        return str(bz2.decompress(buffer), self.get_encoding())



# coding=utf-8
## @package clitellum.endpoints.gateways
#  Este paquete contiene las clases que encapsulan el acceso al sistema de mensajeria con el
# resto de la aplicacion
# Estos gateways contienen amortiguadores, el SenderGateway, antes de enviar un mensaje al sistema
# de mensajeria, lo almacena en una cola en base de datos o bien en fichero. Y el Receiver Gateway
# cuando recibe del sistema de mensajeria, antes de procesarlo, lo almacena en un cola.
# Este tipo de gateways son utiles en la comunicaciones entre los sistemas de mensajeria y sistemas externos.

import threading

from clitellum.core import queue, loadbalancers
from clitellum.core.eventhandling import EventHook
from clitellum.core.fsm import Startable
from clitellum.endpoints.channels import factories
from clitellum.endpoints.channels.events import MessageReceivedArgs


__author__ = 'sergio'


## Clase que expone los metodos basicos para la encapsulacion del sistema de mensajeria
class BaseGatewayQueue(Startable):

    ## Crea una instancia de BasicGateway
    # @param channel Canal de comunicacion utilizado por el gateway
    def __init__(self, queue, channels=list(), numExtractors=4):
        Startable.__init__(self)
        self._channels = list()
        self._queue = queue
        self.OnConnectionError = EventHook()
        for channel in channels:
            self.addChannel(channel)

        self._thExtractors = list()
        for count in range(0, numExtractors):
            self._thExtractors.append(threading.Thread(target=self._process_queue))

    ## Añade un nuevo canal al gateway
    def addChannel(self, channel):
        self._channels.append(channel)
        channel.OnConnectionError += self._invokeOnConnectionError

    ## Realiza la conexion del canal
    def connect(self):
        for channel in self._channels:
            channel.connect()

    ## Cierra la conexion del canal
    def close(self):
        for channel in self._channels:
            channel.close()

    ## Indica si el canal esta o no conectado
    @property
    def is_connected(self):
        for channel in self._channels:
            if channel.is_connected():
                return True

        return False

    def _invokeOnConnectionError(self, sender, args):
        #TODO: Es un error de conexion del canal que hay que tratar ademas de disparar
        self.OnConnectionError.fire(self, args)

    def _process_queue(self):
        pass

    def _invokeOnStart(self):
        Startable._invokeOnStart(self)
        for th in self._thExtractors:
            th.start()

    def _invokeOnStopped(self):
        Startable._invokeOnStopped(self)
        for th in self._thExtractors:
            th.join()

    def __del__(self):
        Startable.__del__(self)
        for channel in self._channels:
            del channel

        self.OnConnectionError.clear()


## Crea un SenderGateway desde un config
# { channels : [
#               { type : 0Mq, timer : Logarithmic, host : tcp://server:8080, maxReconnections : 20 }
#               { type : tcp, timer : Logarithmic, host : tcp://server:8082, maxReconnections : 20 }
#               { type : 0Mq, timer : Logarithmic, host : tcp://server:8083, maxReconnections : 20 }
#               ],
#   router : { type: "RoundRobin" },
#   queue : { type : "Berkeley",  path: "./data/queue.db" }
# }
def CreateSenderFromConfig(config):
    channels = list()
    for ch in config["channels"]:
        channel = factories.CreateOutBoundChannelFromConfig(ch)
        channels.append(channel)

    router = loadbalancers.CreateRouterFromConfig(config.get("balancer"))
    cola = queue.CreateQueueFromConfig(config["queue"])

    return SenderGatewayQueue(router, cola, channels)

## Clase que implementa un gateway de salida
class SenderGatewayQueue (BaseGatewayQueue):
    def __init__(self, loadBalancer, queue, channels=list(), numExtractors=4):
        self._loadBalancer = loadBalancer
        BaseGatewayQueue.__init__(self, queue, channels, numExtractors)

    # TODO: añadir la informacion de enrutamiento, y añadir el canal al router
    def addChannel(self, outBoundChannel):
        BaseGatewayQueue.addChannel(self, outBoundChannel)
        self._loadBalancer.addChannel(outBoundChannel)
        outBoundChannel.OnSendError += self._errorSending
        outBoundChannel.OnMessageSent += self._onMessageSent

    def send(self, message, key):
        self._queue.append([key, message])

    def _process_queue(self):
        while self.state == Startable.RUNNING:
#           TODO: controlar cuando no se puede enviar por uno o por ninguno de los channels
            try:
                item = self._queue.popleft(timeout=10)
                if not item is None:
                    outBoundChannel = self._loadBalancer.next()
                    outBoundChannel.send(item)
                del item
            except Exception:
                pass

    def _errorSending(self, sender, args):
        # Reencolamos el mensaje
        self._queue.task_not_done()

    def _onMessageSent(self, sender, args):
        #Eliminamos el mensaje de la cola definitivamente
        self._queue.task_done()

## Crea un SenderGateway desde un config
# { channels : [
#               { type : 0Mq, timer : Logarithmic, host : tcp://server:8080, maxReconnections : 20 }
#               { type : tcp, timer : Logarithmic, host : tcp://server:8082, maxReconnections : 20 }
#               { type : 0Mq, timer : Logarithmic, host : tcp://server:8083, maxReconnections : 20 }
#               ],
#   numExtractors: 4,
#   queue : { type : "Berkeley",  path: "./data/queue.db" }
# }
def CreateReceiverFromConfig(config):
    channels = list()
    for ch in config["channels"]:
        channel = factories.CreateOutBoundChannelFromConfig(ch)
        channels.append(channel)

    cola = queue.CreateQueueFromConfig(config["queue"])

    return ReceiverGatewayQueue(cola, channels, config['numExtractors'])


## Clase que implementa un gateway de entrada
class ReceiverGatewayQueue(BaseGatewayQueue):

    def __init__(self, queue, channels=list(), numExtractors=4):
        BaseGatewayQueue.__init__(self, queue, channels, numExtractors)
        self.OnMessageReceived = EventHook()

    def addChannel(self, channel):
        BaseGatewayQueue.addChannel(self, channel)
        channel.OnMessageReceived += self._messageReceivedChannel

    def _process_queue(self):
        while self.state == Startable.RUNNING:
            item = self._queue.popleft(timeout=10)
            if not item is None:
                #TODO: Crear los hilos del demonio con el control de errores
                pass

    def _messageReceivedChannel(self, sender, args):
        self._queue.append(args.message)

    def __del__(self):
        BaseGatewayQueue.__del__(self)
        self.OnMessageReceived.clear()

    def _invokeOnReceivedMessage(self, message):
        args = MessageReceivedArgs(message= message)
        self.OnMessageReceived.fire(self, args)




from clitellum.processors.agents import AgentProcessor
from clitellum.endpoints import gateways
from clitellum.endpoints.gateways import SenderGateway, ReceiverGateway

def create_agent_from_config(identification, cfg):
    '''
    Crea un agente desde la configuracion
    Ej:
    {
      input_gateway : ...
      output_gateway : ...
    }
    '''

    builder = AgentProcessorBuilder.create()
    builder.set_identification(identification)
    builder.set_receiver(gateways.CreateReceiverFromConfig(cfg["receiver_gateway"]))
    builder.set_sender(gateways.CreateSenderFromConfig(cfg["sender_gateway"]))

    if not cfg.get("error_gateway") is None:
        builder.set_error_gateway(gateways.CreateSenderFromConfig(cfg["error_gateway"]))

    if 'reply' in cfg:
        if 'sender_gateway' in cfg['reply']:
            builder.set_reply_sender(gateways.CreateSenderFromConfig(cfg['reply']["sender_gateway"]))

        if 'receiver_gateway' in cfg['reply']:
            builder.set_reply_receiver(gateways.CreateReceiverFromConfig(cfg['reply']["receiver_gateway"]))     

        if 'routing_key' in cfg['reply']:
            builder.set_reply_routing_key(cfg['reply']['routing_key'])         

    return builder.build()

class AgentProcessorBuilder(object):

    @classmethod
    def create(cls):
        return AgentProcessorBuilder()

    def __init__(self):
        self._identification = None
        self._receiver_gateway = None
        self._sender_gateway = None
        self._error_gateway = None
        self._reply_sender_gateway = None
        self._reply_receiver_gateway = None
        self._reply_routing_key = None

    def set_identification(self, identification):
        self._identification = identification
        return self

    def set_receiver(self, receiver_gateway):
        self._receiver_gateway = receiver_gateway
        return self

    def set_sender(self, sender_gateway):
        self._sender_gateway = sender_gateway
        return self

    def set_error_gateway(self, error_gateway):
        self._error_gateway = error_gateway
        return self

    def set_reply_sender(self, sender_gateway):
        self._reply_sender_gateway = sender_gateway
        return self

    def set_reply_receiver(self, receiver_gateway):
        self._reply_receiver_gateway = receiver_gateway
        return self

    def set_reply_routing_key(self, routing_key):
        self._reply_routing_key = routing_key
        return self

    def build(self):
        if self._reply_routing_key is not None:
            self._identification.set_reply_key(self._reply_routing_key)

        ag = AgentProcessor(self._identification, self._receiver_gateway, \
                            self._sender_gateway, self._error_gateway)
        if self._reply_sender_gateway is not None:
            ag.set_reply_sender(self._reply_sender_gateway)
        if self._reply_receiver_gateway is not None:
            ag.set_reply_receiver(self._reply_receiver_gateway)

        return ag


# coding=utf-8
import threading

from clitellum.core import loadbalancers, loggerManager
from clitellum.core.eventhandling import EventHook
from clitellum.core.fsm import Startable
from clitellum.endpoints.channels import factories
from clitellum.endpoints.channels.events import MessageReceivedArgs

__author__ = 'sergio'


class BaseGateway(Startable):
    ## Crea una instancia de BasicGateway
    # @param channel Canal de comunicacion utilizado por el gateway
    def __init__(self, channels=list()):
        Startable.__init__(self)
        self._channels = list()
        self.OnConnectionError = EventHook()
        for channel in channels:
            self.addChannel(channel)

    ## Añade un nuevo canal al gateway
    def addChannel(self, channel):
        self._channels.append(channel)
        channel.OnConnectionError += self._invokeOnConnectionError

    ## Realiza la conexion del canal
    def connect(self):
        for channel in self._channels:
            channel.connect()

    ## Cierra la conexion del canal
    def close(self):
        for channel in self._channels:
            channel.close()

    ## Indica si el canal esta o no conectado
    @property
    def is_connected(self):
        for channel in self._channels:
            if channel.is_connected():
                return True

        return False

    def _invokeOnConnectionError(self, sender, args):
        #TODO: Es un error de conexion del canal que hay que tratar ademas de disparar
        self.OnConnectionError.fire(self, args)

    def _process(self):
        pass

    def _invokeOnStart(self):
        Startable._invokeOnStart(self)
        for ch in self._channels:
            ch.start()

    def _invokeOnStopped(self):
        Startable._invokeOnStopped(self)
        for ch in self._channels:
            ch.stop()

    def __del__(self):
        Startable.__del__(self)
        for channel in self._channels:
            del channel

        self.OnConnectionError.clear()


## Crea un SenderGateway desde un config
# { channels : [
#               { type : 0Mq, timer : Logarithmic, host : tcp://server:8080, maxReconnections : 20 }
#               { type : tcp, timer : Logarithmic, host : tcp://server:8082, maxReconnections : 20 }
#               { type : 0Mq, timer : Logarithmic, host : tcp://server:8083, maxReconnections : 20 }
#               ],
#   router : { type: "RoundRobin" }
# }
def CreateSenderFromConfig(config):
    channels = list()
    for ch in config["channels"]:
        if not ch.get('number') is None:
            for index in range(0, ch['number']):
                channel = factories.CreateOutBoundChannelFromConfig(ch)
                channels.append(channel)
        else:
            channel = factories.CreateOutBoundChannelFromConfig(ch)
            channels.append(channel)

    router = loadbalancers.CreateRouterFromConfig(config.get("balancer"))

    return SenderGateway(router, channels)


## Clase que implementa un gateway de salida
class SenderGateway(BaseGateway):
    def __init__(self, loadBalancer, channels=list()):
        self._loadBalancer = loadBalancer
        BaseGateway.__init__(self, channels)

    # TODO: añadir la informacion de enrutamiento, y añadir el canal al router
    def addChannel(self, outBoundChannel):
        BaseGateway.addChannel(self, outBoundChannel)
        self._loadBalancer.addChannel(outBoundChannel)
        outBoundChannel.OnSendError += self._errorSending
        outBoundChannel.OnMessageSent += self._onMessageSent

    def send(self, message, key):
        outBoundChannel = self._loadBalancer.next()
        outBoundChannel.send(message, key)

    def _errorSending(self, sender, args):
        # TODO: Tratamiento de los mensajes y lanzar un evento
        pass

    def _onMessageSent(self, sender, args):
        # TODO: Lanzar un evento
        pass


## Crea un SenderGateway desde un config
# { channels : [
#               { type : 0Mq, timer : Logarithmic, host : tcp://server:8080, maxReconnections : 20 }
#               { type : tcp, timer : Logarithmic, host : tcp://server:8082, maxReconnections : 20 }
#               { type : 0Mq, timer : Logarithmic, host : tcp://server:8083, maxReconnections : 20 }
#               ],
#   numExtractors: 4
# }
def CreateReceiverFromConfig(config):
    channels = list()
    for ch in config["channels"]:
        if not ch.get('number') is None:
            for index in range(0, ch['number']):
                channel = factories.CreateInBoundChannelFromConfig(ch)
                channels.append(channel)
        else:
            channel = factories.CreateInBoundChannelFromConfig(ch)
            channels.append(channel)

    numThreads = 5
    if 'numThreads' in config:
        numThreads = config['numThreads']

    return ReceiverGateway(channels, numThreads)


## Clase que implementa un gateway de entrada
class ReceiverGateway(BaseGateway):
    ## Crea un objeto ReceiverGateway
    # @param channels Lista de canales de recepcion
    # @param numThreads Numero de hilos que procesan los mensajes de los canales, por defecto 1, lo que indica
    # que se usa el mismo hilo de utilizado en la recepción del mensaje
    def __init__(self, channels=list(), numThreads=1):
        BaseGateway.__init__(self, channels)
        self.OnMessageReceived = EventHook()
        self._semaphore = threading.Semaphore(numThreads)
        self._num_threads = numThreads

    def addChannel(self, channel):
        BaseGateway.addChannel(self, channel)
        channel.OnMessageReceived += self._messageReceivedChannel

    def _messageReceivedChannel(self, sender, args):
        if self._num_threads > 1:
            self._semaphore.acquire()
            threading.Thread(target=self.__processMessage, kwargs={"message": args.message}).start()
        else:
            self.__processMessage(args.message)

    def __processMessage(self, message):
        try:
            args = MessageReceivedArgs(message=message)
            self.OnMessageReceived.fire(self, args)
        except Exception as ex:
            loggerManager.get_endPoints_logger().exception("Error al procesar el mensaje")
            raise ex
        finally:
            if self._num_threads > 1:
                self._semaphore.release()

    def __del__(self):
        BaseGateway.__del__(self)
        self.OnMessageReceived.clear()
        del self._semaphore





## @package pylightesb.endpoints.channels.zeromq
#  Este paquete contiene las clases para la comunicacion de tipo zeromq
import zmq
from clitellum.endpoints.channels.basechannels import *


## Clase que establece una conexion de salida de tipo ZeroMq con el host
class OutBoundChannelZeroMq(OutBoundChannel):

    ## Crea una instancia de OutBoundChannelZeroMq
    # @param reconnectionTimer Temporizador de reconexion
    # @param maxReconnections Numero maximo de reconexiones
    # @param host Nombre del host
    def __init__(self, host="", reconnectionTimer=reconnectiontimers.CreateLogarithmicTimer(),
                 maxReconnections=Channel.MAX_RECONNECTIONS, compressor = compressors.DefaultCompressor()):
        self._context = zmq.Context()
        self._socket = self._context.socket(zmq.PUSH)
#        self._socket.setsockopt(28, 10)

        OutBoundChannel.__init__(self, host, reconnectionTimer, maxReconnections,  compressor = compressor)

    def _connect_point(self):
        self._socket.connect(self._host)

    def _close_point(self):
        self._socket.close()
        self._context.destroy()

    def _send(self, message, routingKey='', content_encoding='utf-8', content_type='text/plain'):
        self._socket.send(message)


## Clase que establece una conexion de entrada de tipo ZeroMq con el host
class InBoundChannelZeroMq(InBoundChannel):
    ## Crea una instancia de InBoundChannelZeroMq
    # @param reconnectionTimer Temporizador de reconexion
    # @param maxReconnections Numero maximo de reconexiones
    # @param host Nombre del host
    # @param receptionTimeout Timeout de recepcion de mensaje en milisegudos por defecto 20000
    def __init__(self, host="", reconnectionTimer=reconnectiontimers.CreateLogarithmicTimer(),
                 maxReconnections=Channel.MAX_RECONNECTIONS, receptionTimeout = 20000,  compressor = compressors.DefaultCompressor()):
        InBoundChannel.__init__(self, host, reconnectionTimer, maxReconnections, compressor = compressor)
        self._context = zmq.Context()
        self._socket = self._context.socket(zmq.PULL)
        self._poller = zmq.Poller()
        self._poller.register(self._socket, zmq.POLLIN)
        self._receptionTimeout = receptionTimeout

    def _connect_point(self):
        self._socket.bind(self._host)

    def _close_point(self):
        self._poller.unregister(self._socket)
        self._socket.close()
        self._context.destroy()

    def _startReceive(self):
        evts = self._poller.poll(self._receptionTimeout)
        if len(evts) > 0:
            message = self._socket.recv()
            self._processMessage(message, self._socket)
            del message



from clitellum import errorGateway
from clitellum.core import serialization, loggerManager
from clitellum.core.bus import MessageBuilder
from clitellum.core.fsm import Startable
from clitellum.endpoints import gateways
from clitellum.endpoints.gateways import SenderGateway, ReceiverGateway

__author__ = 'Sergio'


'''
@package clitellum.processors
Este paquete contiene las clases que encapsulan la relacion del los mensajes con sus manejadores
'''


class Bus:
    '''
    Clase que representa la interfaz para poder comunicar con el BUS
    '''
    def __init__(self):
        pass

    @property
    def identification(self):
        return None

    def send(self, message, key, header=None):
        pass

    def reply(self, message, key, header=None):
        pass



class AgentProcessor(Startable, Bus):
    '''
    Clase que implementa el procesamiento de los mensajes
    '''
    def __init__(self, identification, receiver_gateway, sender_gateway, error_gateway=None):
        Startable.__init__(self)
        Bus.__init__(self)
        self._receiver_gateway = receiver_gateway
        self._senderGateway = sender_gateway
        self._identification = identification
        self._receiver_gateway.OnMessageReceived.add(self.__on_message_received)
        self._handler_manager = None
        self._error_gateway = error_gateway
        self._reply_sender = None
        self._reply_receiver = None

    @property
    def identification(self):
        return self._identification

    def set_reply_sender(self, sender_gateway: SenderGateway):
        self._reply_sender = sender_gateway

    def set_reply_receiver(self, receiver_gateway: ReceiverGateway):
        self._reply_receiver = receiver_gateway
        self._reply_receiver.OnMessageReceived.add(self.__on_message_received)

    def configure(self, handler_manager):
        self._handler_manager = handler_manager

    def __on_message_received(self, sender, args):
        try:
            message_bus = serialization.loads(args.message)

            handler = self._handler_manager.get_handler(message_bus['Header']['BodyType'])
            handler.initialize(self, message_bus['Header'])

            handler.handle_message(message_bus['Body'])

        except Exception as ex:
            loggerManager.get_processors_logger().exception("Error al procesar el mensaje %s", args.message)
            self._send_error(args.message, ex)

    def _invokeOnStart(self):
        Startable._invokeOnStart(self)
        self._receiver_gateway.start()
        self._senderGateway.connect()
        if self._error_gateway is not None:
            self._error_gateway.connect()
        
        if self._reply_sender is not None:
            self._reply_sender.connect()
        
        if self._reply_receiver is not None:
            self._reply_receiver.start()

    def _invokeOnStopped(self):
        Startable._invokeOnStopped(self)
        self._receiver_gateway.stop()
        self._senderGateway.close()
        if self._error_gateway is not None:
            self._error_gateway.close()

        if self._reply_sender is not None:
            self._reply_sender.close()
        
        if self._reply_receiver is not None:
            self._reply_receiver.stop()


    def send(self, message, key, header=None):

        message_bus = MessageBuilder.create() \
                      .set_service(self.identification) \
                      .set_type(key) \
                      .set_body(message) \
                      .set_context(header) \
                      .build()

        message_str = serialization.dumps(message_bus)
        self._senderGateway.send(message_str, key)

    def reply(self, message, key, header=None):        
        reply_service = header['CallStack'].pop()
        message_bus = MessageBuilder.create() \
                      .set_service(self.identification) \
                      .set_type(key) \
                      .set_body(message) \
                      .set_context(header) \
                      .set_reply() \
                      .build()

        message_str = serialization.dumps(message_bus)
        if self._reply_sender is None:
            self._senderGateway.send(message_str, key)
        else:
            self._reply_sender.send(message_str, reply_service['ReplyKey'])


    def _send_error(self, message_received, exception):
        if self._error_gateway is None:
            return
        message = errorGateway.create_error_message(message_received, exception)
        message_bus = MessageBuilder.create() \
                      .set_service(self.identification) \
                      .set_type('Error.ErrorHandler') \
                      .set_body(message) \
                      .build()

        message_str = serialization.dumps(message_bus)
        self._error_gateway.send(message_str, "Error.ErrorHandler")

    def __del__(self):
        Startable.__del__(self)



# author = sbermudel
# package description
from datetime import date
import datetime
import pytz

## Clase que contiene la informacion de identificacion de un proceso
class Identification:
    '''
    Contiene la información de identificacion del servicio
    '''

    def __init__(self, id_service, type_service):
        self._id = id_service
        self._type = type_service
        self._reply_key = type_service

    @property
    def id(self):
        '''
        Devuelve el identificador del servicio
        '''
        return self._id

    @property
    def type(self):
        '''
        Devuelve el tipo del servicio
        '''
        return self._type

    def set_reply_key(self, routing_key):
        '''
        Estable la clave de enrutamiento para los mensajes de reply
        '''
        self._reply_key = routing_key

    def get_reply_key(self):
        '''
        Devuelve la clave de enrutamiento para los mensajes de reply
        '''
        return self._reply_key

class MessageBuilder:
    '''
    Constructor de mensajes
    '''
    def __init__(self):
        self._identifcation = None
        self._body = None
        self._body_type = None
        self._call_context = dict()
        self._stack_trace = list()
        self._call_stack = list()
        self._is_reply = False
        self._priority = 0

    @classmethod
    def create(cls):
        '''
        Devuelve un builder de Message
        '''
        return MessageBuilder()

    def set_service(self, identification):
        '''
        Establece el servicio que crea el mensaje
        '''
        self._identifcation = identification
        return self

    def set_body(self, body):
        '''
        Establece el body
        '''
        self._body = body
        return self

    def set_type(self, body_type):
        '''
        Establece el tipo de mensaje
        '''
        self._body_type = body_type
        return self
    
    def set_context(self, header):
        '''
        Establece el contexto de ejecucion del mensaje
        '''
        if header is None:
            return self

        if 'CallContext' in header:
            self._call_context = header['CallContext']

        if 'CallStack' in header:
            self._call_stack = header['CallStack']

        if 'StackTrace' in header:
            self._stack_trace = header['StackTrace']

        return self

    def set_reply(self):
        '''
        Establece que el mensaje es un reply
        '''
        self._is_reply = True
        return self

    def set_priority(self, priority):
        '''
        Estable que la prioridad del mensaje por defecto es cero
        '''
        self._priority = priority
        return self

    def build(self):
        '''
        Contruye el mensaje
        '''
        message = dict()
        message['Body'] = self._body
        message['Header'] = self._generate_header()
        return message

    def _generate_header(self):
        header = dict()
        header['BodyType'] = self._body_type
        header['Priority'] = self._priority
        header['CreatedAt'] = datetime.datetime.utcnow().replace(tzinfo=pytz.timezone("UTC")).isoformat()
        header['CallContext'] = self._call_context
        header['CallStack'] = self._call_stack
        header['StackTrace'] = self._stack_trace
        header['IdentificationService'] = self._generate_identificaction()
        header['IsReply'] = self._is_reply

        header['StackTrace'].append(
            { 'Service' : header['IdentificationService'], 'CallContext': self._call_context })

        if not self._is_reply:
            header['CallStack'].append(header['IdentificationService'])

        return header

    def _generate_identificaction(self):
        identification = dict()
        identification['Id'] = self._identifcation.id
        identification['Type'] = self._identifcation.type
        identification['ReplyKey'] = self._identifcation.get_reply_key()
        return identification



# author = sbermudel
# package description
import re
import socket
import struct
import threading
from time import sleep

from clitellum.core import compressors, loggerManager
from clitellum.endpoints.channels import reconnectiontimers
from clitellum.endpoints.channels.basechannels import OutBoundChannel, Channel, InBoundChannel
from clitellum.endpoints.channels.exceptions import ConnectionError, SendError


class BaseChannelTcp:
    Ack_Message = "Ack"

    ## Crea una instancia de OutBoundChannelTcp
    # @param reconnectionTimer Temporizador de reconexion
    # @param maxReconnections Numero maximo de reconexiones
    # @param host Nombre del host ej: tcp://server:port
    def __init__(self, host=""):
        match = re.search('^(\w+)://(.*):(\d+)', host)
        if match:
            self._protocol = match.group(1)
            self._server = match.group(2)
            self._port = int(match.group(3))
        else:
            raise NameError("Invalid host name", host)

        self._address = socket.getaddrinfo(self._server, self._port)

        self._options = socket.IPPROTO_TCP
        if self._protocol == "udp":
            self._options = socket.IPPROTO_UDP

    def _create_socket(self):
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM, proto=self._options)
        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)

    def _close_point(self):
        try:
            self._socket.shutdown(socket.SHUT_RDWR)
            self._socket.close()
            sleep(5)
            self._socket = None
        except Exception as ex:
            loggerManager.get_endPoints_logger().error("Error al desconectar del host %s:%i : %s" % (self._server, self._port, str(ex)))
        finally:
            del self._socket


class OutBoundChannelTcp(OutBoundChannel, BaseChannelTcp):
    ## Crea una instancia de OutBoundChannelTcp
    # @param reconnectionTimer Temporizador de reconexion
    # @param maxReconnections Numero maximo de reconexiones
    # @param host Nombre del host ej: tcp://server:port
    def __init__(self, host="", reconnectionTimer=reconnectiontimers.CreateLogarithmicTimer(),
                 maxReconnections=Channel.MAX_RECONNECTIONS, compressor=compressors.DefaultCompressor(), useAck=False):
        OutBoundChannel.__init__(self, host, reconnectionTimer, maxReconnections,
                                 compressor=compressor, useAck=useAck)
        BaseChannelTcp.__init__(self, host)

    def _connect_point(self):
        try:
            self._create_socket()
            self._socket.connect((self._server, self._port))
        except socket.error as msg:
            raise ConnectionError(msg)

    def _close_point(self):
        BaseChannelTcp._close_point(self)

    def _send(self, message, content_encoding='utf-8', content_type='text/plain'):
        buff = ''
        try:
            buff = struct.pack('!I%ds' % (len(message),), len(message), message)
            self._socket.send(buff)
        except socket.error as ex:
            raise SendError("Error al enviar el mensaje %s: %s" % (message, ex))
        finally:
            del buff

    def _waitForAck(self):
        try:
            message = ''
            bytesReceived = self._socket.recv(4)
            if not len(bytesReceived) == 4:
                return False

            messageLength = struct.unpack('>I', bytesReceived)[0]
            while len(message) < messageLength:
                chunk = self._socket.recv(messageLength)
                message += chunk

            del bytesReceived, messageLength
            return message == BaseChannelTcp.Ack_Message

        except:
            return False

## Clase que establece una conexion de entrada de tipo TCP con el host
class InBoundChannelTcp(InBoundChannel, BaseChannelTcp):
    ## Crea una instancia de InBoundChannelTcp
    # @param reconnectionTimer Temporizador de reconexion
    # @param maxReconnections Numero maximo de reconexiones
    # @param host Nombre del host
    # @param receptionTimeout Timeout de recepcion de mensaje en milisegudos por defecto 20000
    def __init__(self, host="", reconnectionTimer=reconnectiontimers.CreateLogarithmicTimer(),
                 maxReconnections=Channel.MAX_RECONNECTIONS, receptionTimeout=10,
                 compressor=compressors.DefaultCompressor(),
                 maxConnectionRequests=100, useAck=False):
        InBoundChannel.__init__(self, host, reconnectionTimer, maxReconnections, compressor=compressor,
                                useAck=useAck)
        BaseChannelTcp.__init__(self, host)
        self._maxConnectionRequests = maxConnectionRequests
        self.__receptionTimeout = receptionTimeout
        self.__thReadData = list()

    def _create_socket(self):
        BaseChannelTcp._create_socket(self)
        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._socket.settimeout(self.__receptionTimeout)
        socket.setdefaulttimeout(self.__receptionTimeout)

    def _connect_point(self):
        try:
            self._create_socket()
            self._socket.bind((self._server, self._port))
            self._socket.listen(self._maxConnectionRequests)
        except Exception as ex:
            loggerManager.get_endPoints_logger().error(
                "Error al conectar con el host %s:%i" % (self._server, self._port))
            loggerManager.get_endPoints_logger().error("Socket de mierda: %s" % self._socket.fileno)
            raise ConnectionError(ex)

    def _close_point(self):
        BaseChannelTcp._close_point(self)

    def _startReceive(self):
        try:
            loggerManager.get_endPoints_logger().debug("Aceptando conexiones")
            (clientSocket, address) = self._socket.accept()
            loggerManager.get_endPoints_logger().info("Conexion establecida con el cliente %s:%i." % (address[0], address[1]))
            th = threading.Thread(target=InBoundChannelTcp.__readData, args=(self, clientSocket), name="TCP client %s:%i" % (address[0], address[1]))
            loggerManager.get_endPoints_logger().debug('Iniciando hilo de lectura %s' % th.getName())
            th.start()
        except socket.timeout as ex:
            loggerManager.get_endPoints_logger().debug("Timeout de aceptacion de conexiones : %s." % ex.message)
            pass
        except Exception as ex:
            loggerManager.get_endPoints_logger().error("Error al aceptar conexiones : %s." % str(ex))
            raise ex

    def __readData(self, clientSocket):
        clientSocket.settimeout(self.__receptionTimeout)
        while self.isRunning:
            try:
                loggerManager.get_endPoints_logger().debug("Leyendo mensaje...")
                message = ''
                idMessage = ''
                messageLength = 0
                bytesReceived = ''
                bytesReceived = clientSocket.recv(4)
                if not len(bytesReceived) == 4:
                    raise Exception("Error al leer cabecera")
                loggerManager.get_endPoints_logger().debug("Cabecera leida.")
                messageLength = struct.unpack('>I', bytesReceived)[0]
                message = clientSocket.recv(messageLength)
                if not len(message) == messageLength:
                    raise Exception("Error de longitud del mensaje : Expected length %s - Actual length %i" % (messageLength, len(message)))
                idMessage = clientSocket.recv(36)
                if not len(idMessage) == 36:
                    raise Exception("Error de identificacion del mensaje : Msg length %i - Msg Id length %i" % (len(message), len(idMessage)))
                self._processMessage(message, idMessage, clientSocket)
            except socket.timeout as ex:
                loggerManager.get_endPoints_logger().error("Error de recepcion : %s." % ex.message)
                break
            except socket.error as ex:
                loggerManager.get_endPoints_logger().error("Error de socket : %s." % str(ex))
                break
            except Exception as ex:
                loggerManager.get_endPoints_logger().error(str(ex))
                break
            finally:
                del message, bytesReceived, messageLength, idMessage
        try:
            clientSocket.shutdown(socket.SHUT_RDWR)
            clientSocket.close()
            clientSocket = None
        except Exception as ex:
            loggerManager.get_endPoints_logger().error("Error al desconectar del cliente : %s" % ex.message)
        finally:
            del clientSocket

    def _sendAck(self, object, idMessage):
        buff = struct.pack('!I%ds' % (len(idMessage),), len(idMessage), idMessage)
        try:
            object.send(buff)
        except Exception as ex:
            loggerManager.get_endPoints_logger().error("Error al enviar el ACK %s" % str(ex))
            raise ex
        finally:
            del buff





# author = sbermudel
## @package clitellum.core.fsm
#  Este paquete contiene las clases de maquinas de estado
from fluidity import *

from clitellum.core.eventhandling import EventHook



## Implemeta la una maquina de estados start-stop
class StartableStateMachine(StateMachine):

    initial_state = 'Stopped'

    state('Running', enter='_invokeOnStart')
    state('Stopped', enter='_invokeOnStopped')

    transition(from_=['Running','Stopped'], event='start', to='Running')
    transition(from_=['Running','Stopped'], event='stop', to='Stopped')


    ## Crea una isntancia de StartableStateMachine
    def __init__(self):
        self.OnStart = EventHook()
        self.OnStopped = EventHook()
        StateMachine.__init__(self)

    ## Lanza el evento OnStart
    def _invokeOnStart(self):
        self.OnStart.fire()

    ## Lanza el evento OnStopped
    def _invokeOnStopped(self):
        self.OnStopped.fire()

    ## Destructor
    def __del__(self):
        self.OnStart.clear()
        self.OnStopped.clear()


## Esta clase encapsula una maquina de estado capaz de iniciar y parar un proceso
class Startable():

    RUNNING = "Running"
    STOPPED = "Stopped"

    def __init__(self):
        self._statemachine = StartableStateMachine()
        self.OnStart = EventHook()
        self.OnStopped = EventHook()
        self._statemachine.OnStart += self._invokeOnStart
        self._statemachine.OnStopped += self._invokeOnStopped

    ## Devuelve el estado acutal del proceso
    @property
    def state(self):
        return self._statemachine.current_state

    ## Incia el proceso
    def start(self):
        self._statemachine.start()

    ## Termina el proceso
    def stop(self):
        self._statemachine.stop()

    ## Lanza el evento OnStart
    def _invokeOnStart(self):
        self.OnStart.fire()

    ## Lanza el evento OnStopped
    def _invokeOnStopped(self):
        self.OnStopped.fire()

    ## Destructor
    def __del__(self):
        self.OnStart.clear()
        self.OnStopped.clear()


# coding=utf-8
## @package clitellum.core.eventhandling
#  Este paquete contiene las clases para la gestion de eventos
__author__ = 'sergio'


## Clase que implementa event handling
class EventHook(object):
    ## Crea una instancia del contralador de eventos
    #
    def __init__(self):
        self.__handlers = []

    ## Permite añadir un handler al controlador de eventos
    # @param handler Funcion a la que se llama cuando se lance el evento
    def __iadd__(self, handler):
        self.__handlers.append(handler)
        return self

    ## Elimina un handler del controlador de eventos
    # @param handler Funcion a la que se llama cuando se lance el evento
    def __isub__(self, handler):
        self.__handlers.remove(handler)
        return self

    ## Permite añadir un handler al controlador de eventos
    # @param handler Funcion a la que se llama cuando se lance el evento
    def add(self, handler):
        self.__handlers.append(handler)
        return self

    ## Elimina un handler del controlador de eventos
    # @param handler Funcion a la que se llama cuando se lance el evento
    def remove(self, handler):
        self.__handlers.remove(handler)
        return self

    ## Dispara todos los handelrs asociados
    # @param args Argumentos del evento
    # @param keywargs Argumentos del evento
    def fire(self, *args, **keywargs):
        for handler in self.__handlers:
            handler(*args, **keywargs)

    ## Elimina un handler especificado del contraldor de eventos
    # @param inObject Handler que se quiere eliminar
    def clearObjectHandlers(self, inObject):
        for theHandler in self.__handlers:
            if theHandler.im_self == inObject:
                self -= theHandler

    ## Elimina todos los handlers del controlador de eventos
    def clear(self):
        for theHandler in self.__handlers:
            self -= theHandler

    ## Destructor de clase, elimina todos los handlers del controlador de eventos
    #
    def __del__(self):
        self.clear()




from time import sleep
import math
from clitellum.core import loggerManager

__author__ = 'sergio'
## @package clitellum.endpoints.channels.reconnectiontimers
#  Este paquete contiene las clases para los temporizadores de reconexion
#


## Metodo factoria que crea una instancia de un temporizador
# instantaneo
def CreateInstantTimer():
    return InstantReconnectionTimer()


## Metodo factoria que crea una instancia de un temporizador
# logaritmico
def CreateLogarithmicTimer():
    return LogarithmicReconnectionTimer()

## Metodo factoria que crear una instancia de un temporizador de tiempo constante
def CreateConstantTimer(waiting_time=5):
    return ConstantReconnectionTimer(waiting_time=waiting_time)


## Crea una temporizador en funcion del tipo especificado
# @param type Tipo de temporizador "Instant", "Logarithmic"
def CreateTimerFormType(type):
    if type == "Instant":
        return CreateInstantTimer()
    elif type == 'Constant':
        return ConstantReconnectionTimer()
    else:
        return CreateLogarithmicTimer()

## Crea un temporizador a partir de una configuracion
# { type :'Instant' }
# { type :'Constant', time : 10 }
# { type :'Logarithmic' }
def CreateTimerFormConfig(config):
    if config['type'] == "Instant":
        return CreateInstantTimer()
    elif config['type'] == 'Constant':
        if not config.get['time'] is None:
            return ConstantReconnectionTimer(config['time'])
        else:
            return ConstantReconnectionTimer()
    else:
        return CreateLogarithmicTimer()



## Clase base que proporciona la estructura basica de un temporizador de reconexion
class ReconnectionTimer:
    ## Crea una instancia del temporizador de reconexion
    def __init__(self):
        pass

    ## Se espera una vuelta del ciclo antes de continuar
    def wait(self):
        pass

    ## Reinicia el temporizador
    def reset(self):
        pass


## Clase que proporciona un temporizador de reconexion instantaneo,
# no hay tiempo de espera entre un ciclo y el siguiente
class InstantReconnectionTimer(ReconnectionTimer):
    ## Crea una instancia del temporizador instantaneo
    def __init__(self):
        ReconnectionTimer.__init__(self)

    ## Convierte la instancia a string
    def __str__(self):
        return "Instant Reconnection Timer"


## Define un temporizador de reconexion en el que el tiempo de espera entre un ciclo
# y el siguiente es logaritmico, .
class LogarithmicReconnectionTimer(ReconnectionTimer):
    def __init__(self):
        ReconnectionTimer.__init__(self)
        self._seed = 1

    def wait(self):
        try:
            waitingTime = ((1 + (1 / self._seed)) ** self._seed) * (1 + math.log10(self._seed))
            if waitingTime < 0:
                waitingTime = 0
            sleep(waitingTime)
            self._seed += 1
        except Exception as ex:
            loggerManager.get_endPoints_logger().error("Error en el temporizador logaritmico %s" % ex)

    def reset(self):
        self._seed = 1

    ## Convierte la instancia a string
    def __str__(self):
        return "Logarithmic Reconnection Timer, seed: %s" % self._seed


## Define un temporizador de reconexion en el que el tiempo de espera entre un ciclo
# y el siguiente es logaritmico, .
class ConstantReconnectionTimer(ReconnectionTimer):
    def __init__(self, waiting_time=5):
        ReconnectionTimer.__init__(self)
        self._waiting_time = waiting_time

    def wait(self):
        sleep(self._waiting_time)

    def reset(self):
        pass

    ## Convierte la instancia a string
    def __str__(self):
        return "Constant Reconnection Timer, seed: %s" % self._waiting_time



import re
import socket
import threading

import pika

from clitellum.core import compressors, loggerManager
from clitellum.endpoints.channels import reconnectiontimers
from clitellum.endpoints.channels.basechannels import OutBoundChannel, Channel, InBoundChannel
from clitellum.endpoints.channels.exceptions import ConnectionError
from clitellum.endpoints.channels.amqp.configuration import AmqpHostConfiguration

__author__ = 'sergio'


class BaseAmqpChannel(object):
    '''
    Clase base de un canal amqp
    '''

    def __init__(self, amqp_config: AmqpHostConfiguration):
        self._amqp_config = amqp_config
        self._connection = None
        self._channel = None

    def _connect_point(self):
        try:
            if not self._connection is None and self._connection.is_open:
                self._connection.close()

            parameters = pika.connection.Parameters()
            parameters.host = self._amqp_config.get_server()
            parameters.heartbeat = 60
            parameters.port = self._amqp_config.get_port()
            parameters.credentials = pika.PlainCredentials(username=self._amqp_config.get_user(), \
                                                        password=self._amqp_config.get_password())

            self._connection = pika.BlockingConnection(parameters=parameters)
            self._channel = self._connection.channel()

            if self._amqp_config.get_exchange().must_be_created():
                self._channel.exchange_declare(exchange=self._amqp_config.get_exchange().get_name(), \
                durable=True, exchange_type=self._amqp_config.get_exchange().get_type(), \
                auto_delete=False)

        except Exception as ex:
            raise ConnectionError(ex)

    def _close_point(self):
        if self._connection.is_open():
            self._connection.close()


class OutBoundAmqpChannel(OutBoundChannel, BaseAmqpChannel):
    '''
    Clase que implementa un canal de salida con el protocolo amqp
    '''

    def __init__(self, amqp_config: AmqpHostConfiguration, \
                reconnectionTimer=reconnectiontimers.CreateLogarithmicTimer(), 
                maxReconnections=Channel.MAX_RECONNECTIONS, compressor=compressors.DefaultCompressor()):


        BaseAmqpChannel.__init__(self, amqp_config)
        OutBoundChannel.__init__(self, self._amqp_config.get_exchange().get_name(), \
                                 reconnectionTimer, maxReconnections, \
                                 compressor, useAck=amqp_config.get_use_ack())

    def _connect_point(self):
        BaseAmqpChannel._connect_point(self)
        if self._useAck:
            self._channel.confirm_delivery()

    def _close_point(self):
        BaseAmqpChannel._close_point(self)

    def _send(self, message, routing_key='', content_encoding='utf-8', content_type='text/plain'):
        try:
            properties = pika.BasicProperties(content_type=content_type, delivery_mode=2, content_encoding=content_encoding)
            if routing_key == '':
                return self._channel.basic_publish(body=message,
                                                   exchange=self._amqp_config.get_exchange().get_name(),
                                                   routing_key='',
                                                   properties=properties)
            else:
                return self._channel.basic_publish(body=message,
                                                   exchange=self._amqp_config.get_exchange().get_name(),
                                                   routing_key=routing_key,
                                                   properties=properties)

        except socket.error as ex:
            loggerManager.get_endPoints_logger().error("Error: %s" % ex)
            raise ConnectionError("Se ha perdido la conexcion con el servidor AMPQ")
        except Exception as ex:
            loggerManager.get_endPoints_logger().error("Error: %s" % ex)
            raise ConnectionError('Error al enviar el elemento %s' % ex)

class InBoundAmqpChannel(InBoundChannel, BaseAmqpChannel):

    def __init__(self, amqp_config: AmqpHostConfiguration, \
                 reconnectionTimer=reconnectiontimers.CreateLogarithmicTimer(),
                 maxReconnections=Channel.MAX_RECONNECTIONS, receptionTimeout=10,
                 compressor=compressors.DefaultCompressor(),
                 max_threads=1):

        BaseAmqpChannel.__init__(self, amqp_config)
        InBoundChannel.__init__(self, self._amqp_config.get_exchange().get_name(), \
                                reconnectionTimer, maxReconnections, compressor=compressor,
                                useAck=self._amqp_config.get_use_ack())

        self._reception_timeout = receptionTimeout
        self.__is_consuming = False
        self.__max_threads = max_threads
        self.__semaforo = threading.Semaphore(max_threads)
        self.__ack_semaforo = threading.Semaphore()
        self.__consumer_tag = None


    def _connect_point(self):
        BaseAmqpChannel._connect_point(self)

        try:
            if self._amqp_config.get_queue().must_be_created():
                self._channel.queue_declare(queue=self._amqp_config.get_queue().get_name(), \
                                            durable=True, auto_delete=False)
                for key in self._amqp_config.get_queue().get_routing_keys():
                    self._channel.queue_bind(queue=self._amqp_config.get_queue().get_name(), \
                                             exchange=self._amqp_config.get_exchange().get_name(), \
                                             routing_key=key)

            self._channel.basic_qos(prefetch_size=self._amqp_config.get_qos().get_prefetch_size(),
                                    prefetch_count=self._amqp_config.get_qos().get_prefetch_count())

        except Exception as ex:
            raise ConnectionError(ex)

    def _close_point(self):
        BaseAmqpChannel._close_point(self)

    def __read_message(self, channel, method_frame, header_frame, body):
        if self.__max_threads > 1:
            self.__semaforo.acquire()
            msg = {"body": body, "channel" : channel, "delivery_tag": method_frame.delivery_tag}
            threading.Thread(target=self.__worker, kwargs={"msg": msg}).start()
        else:
            msg_info = {"channel": channel, "delivery_tag": method_frame.delivery_tag}
            self._processMessage(body, 0, msg_info)

    def __worker(self, msg):
        msg_info = {"channel": msg['channel'], "delivery_tag": msg['delivery_tag']}
        self._processMessage(msg['body'], 0, msg_info)
        self.__semaforo.release()

    def _startReceive(self):
        self.__consumer_tag = self._channel.basic_consume(self.__read_message,
                                                          queue=self._amqp_config.get_queue().get_name(),
                                                          no_ack=False)
        self._connection.add_timeout(self._reception_timeout, self._stopReceive)
        self._channel.start_consuming()

    def _wait_method(self):
        pass

    def _stopReceive(self):
        self._channel.basic_cancel(self.__consumer_tag)

    def _sendAck(self, object, idMessage):
        self.__ack_semaforo.acquire()
        object["channel"].basic_ack(delivery_tag=object["delivery_tag"])
        self.__ack_semaforo.release()



## @package pylightesb.endpoints.channels.basechannels
#  Este paquete contiene las clases para la comunicacion entre componentes
import threading

from clitellum.core import compressors, loggerManager
from clitellum.endpoints.channels.events import *
from clitellum.endpoints.channels.exceptions import *
from clitellum.core.eventhandling import EventHook
from clitellum.core.fsm import Startable
from clitellum.endpoints.channels import reconnectiontimers

## Clase base que gestiona que define un punto de conexion con un host, y que gestiona
# las reconexiones mediante un temporizador
class Channel:
    INFINITE_RECONNECTIONS = -1
    MAX_RECONNECTIONS = INFINITE_RECONNECTIONS


    ## Crea una instancia de un punto de conexion, y estable el temporizador
    # de la reconexion
    # @param host Nombre del host
    # @param reconnectionTimer Temporizador de la reconexion (ReconnectionTimer) por defecto Logaritmico
    # @param maxReconnections Numero maximo de reconexiones por defecto 10
    # @param useCompression Indica si el punto de conexion usa compresion
    def __init__(self, host="", reconnectionTimer=reconnectiontimers.CreateLogarithmicTimer(),
                 maxReconnections=MAX_RECONNECTIONS, compressor=compressors.DefaultCompressor(), useAck=False):
        self._isConnected = False
        self._reconnectionTimer = reconnectionTimer
        self._reconnectionNumber = 0
        self._maxReconnections = maxReconnections
        self.OnConnectionError = EventHook()
        self._host = host
        self._useAck = useAck
        self._compressor = compressor

    ## Establece la conexion con el host, en el caso de no establecerse
    # se procede a la reconexion hasta que el temporizador llegue a su limite de intentos
    # de reconexion
    def connect(self):
        try:
            self._reconnect()
        except ConnectionError:
            self._invokeOnConnectionError()

    ## Indica si el punto de conexion esta conectado con el host
    def is_connected(self):
        return self._isConnected

    ## Devuelve el nombre del host del punto de conexion
    def getHost(self):
        return self._host

    ## Cierra la conexion con el host
    def close(self):
        try:
            if self._isConnected:
                self._close_point()
        except:
            pass

        self._isConnected = False


    ## Metodo que realiza la conexion fisica con el host
    # @exception ConnectException Se lanza en el caso de que no sea posible la conexion
    def _connect_point(self):
        raise ConnectionError("Error de conexion, clase base")

    ## Cierra la conexion fisica con el host
    def _close_point(self):
        raise ConnectionError("Error de conexion, clase base")

    ## Realiza la reconexion en el caso de que se produca una excepcion
    # @exception ConnectException Se lanza en el caso de que se haya llegado al limite de
    # intentos de reconexion
    def _reconnect(self):
        # Si ya esta conectado no reconecta
        if self._isConnected:
            return

        self._reconnectionNumber = 0
        self._reconnectionTimer.reset()
        self._isConnected = False

        while True:
            try:
                loggerManager.get_endPoints_logger().debug("Conectando al host: %s", self._host)
                self._connect_point()
                break
            except ConnectionError as ex:
                loggerManager.get_endPoints_logger().error("Error al conectar: %s" % ex)
                if self._maxReconnections != Channel.INFINITE_RECONNECTIONS and self._reconnectionNumber >= self._maxReconnections:
                    loggerManager.get_endPoints_logger().error("Se ha alcanzado el maximo numero de reconexiones")
                    raise ConnectionError("Se ha alcanzado el maximo numero de reconexiones")
                else:
                    loggerManager.get_endPoints_logger().error("Intentado reconectar... %d, Host: %s" % (self._reconnectionNumber, self._host))
                    self._reconnectionNumber += 1
                    self._reconnectionTimer.wait()
        self._isConnected = True

    ## Lanza el evento OnConnectionError
    def _invokeOnConnectionError(self):
        args = ConnectionErrorArgs(channel=self, maxReconnections=self._maxReconnections,
                                   numReconnections=self._reconnectionNumber, timer=self._reconnectionTimer)
        self.OnConnectionError.fire(self, args)
        del args

    ## Destructor, si la conexion esta abierta la cierra y libera todos los handler del
    # controlador de eventos
    def __del__(self):
        if self.is_connected():
            self.close()
        self.OnConnectionError.clear()


## Clase que implementa un un punto de conexion de salida
class OutBoundChannel(Channel):
    ## Crea una instancia de OutBoundChannel
    # @param reconnectionTimer Temporizador de reconexion
    # @param maxReconnections Numero maximo de reconexiones
    # @param host Nombre del host
    def __init__(self, host="", reconnectionTimer=reconnectiontimers.CreateLogarithmicTimer(),
                 maxReconnections=Channel.MAX_RECONNECTIONS, compressor=compressors.DefaultCompressor(), useAck=False):
        Channel.__init__(self, host, reconnectionTimer, maxReconnections, compressor, useAck)
        self.OnSendError = EventHook()
        self._mutex = threading.Lock()
        self.OnMessageSent = EventHook()

    ## Lanza el evento
    def _invokeOnSendError(self, message):
        args = SendErrorArgs(channel=self, message=message)
        self.OnSendError.fire(self, args)
        del args

    ## Lanza el evento de mensaje enviado
    def _invokeOnMessageSent(self, message):
        args = SentMessageArgs(channel=self, message=message)
        self.OnMessageSent.fire(self, args)
        del args


    ## Envia un mensaje al punto de conexion
    def send(self, message, routingKey=''):
        ret = True
        self._mutex.acquire()
        while True:
            try:
                #loggerManager.get_endPoints_logger().debug("Enviando mensaje %s" % message)
                msg = self._compressor.compress(message)
                ret = self._send(msg, routingKey, self._compressor.get_encoding(), 'application/json')
                if not ret:
                    self.__checkMessageSent(message)
                else:
                    self._invokeOnMessageSent(message)

                del msg
                break
            except ConnectionError as ex:
                loggerManager.get_endPoints_logger().error("Error de conexion al enviar el mensaje %s" % ex)
                # Intentamos reconectar
                try:
                    self.close()
                    self._reconnect()
                except ConnectionError as ex2:
                    loggerManager.get_endPoints_logger().error(
                        "No se ha podido reconectar al enviar el mensaje %s" % ex2)
                    self._invokeOnSendError(message)
                    break
            except SendError as ex:
                loggerManager.get_endPoints_logger().error("Error de envio %s" % ex)
                self._invokeOnSendError(message)
                break
            except Exception:
                raise

        self._mutex.release()
        return ret

    ## Envia un mensaje al punto de conexion
    # @param message Buffer a enviar
    # @param routingKey Clave de enrutamiento
    # @exception ConnectionError
    def _send(self, message, routingKey='', content_encoding='utf-8', content_type='text/plain'):
        raise SendError("Metodo no implementado")

    ## Espera el acuse de recibo del mensaje enviado
    def _waitForAck(self):
        return True

    ## Realiza la comprobacion de que el mensaje a sido enviado y lanza los eventos
    def __checkMessageSent(self, message):
        if self._useAck:
            ret = self._waitForAck()
            if ret:
                self._invokeOnMessageSent(message)
            else:
                self._invokeOnSendError(message)
            return ret
        else:
            self._invokeOnMessageSent(message)
            return True

    def __del__(self):
        self.OnMessageSent.clear()
        self.OnSendError.clear()
        Channel.__del__(self)


## Clase que implementa un un punto de conexion de salida
class InBoundChannel(Channel, Startable):
    ## Crea una instancia de InBoundChannel
    # @param reconnectionTimer Temporizador de reconexion
    # @param maxReconnections Numero maximo de reconexiones
    # @param host Nombre del host
    def __init__(self, host="", reconnectionTimer=reconnectiontimers.CreateLogarithmicTimer(),
                 maxReconnections=Channel.MAX_RECONNECTIONS, compressor=compressors.DefaultCompressor(), useAck=False):
        Channel.__init__(self, host, reconnectionTimer, maxReconnections, compressor, useAck)
        Startable.__init__(self)
        self._thReceive = threading.Thread(target=self.__workReceive)
        self.OnMessageReceived = EventHook()
        self.OnStart += self.__beginReceive
        self.OnStopped += self.__endReceive

    ## Lanza el evento Message Received
    def __invokeOnMessageReceived(self, message):
        args = MessageReceivedArgs(message=message)
        self.OnMessageReceived.fire(self, args)
        del args

    ## Lanza el inicio de la recepcion de mensajes
    def __beginReceive(self):
        if not self._isConnected:
            self.connect()
        self._initializeReceive()
        self._thReceive.start()

    ## Propiedad que indica si el punto de conexion esta recibiendo mensajes
    @property
    def isRunning(self):
        return self.state == Startable.RUNNING

    ## Metodo del hilo en el que se ejecuta la recepcion de mensajes
    def __workReceive(self):
        while True:
            try:
                if self.isRunning:
                    self._startReceive()
                else:
                    self._stopReceive()
                    break
            except Exception as ex:
                # Intentamos reconectar
                try:
                    loggerManager.get_endPoints_logger().error("Error en la recepcion de mensajes %s" % ex)
                    self.close()
                    self._reconnect()
                except ConnectionError:
                    self.stop()
                    self._invokeOnConnectionError()
                    break

    ## Finaliza la recepcion de mensajes
    def __endReceive(self):
        self._thReceive.join()
        self._terminateReceive()

    ## Metodo al que las clases hijas deben llamar cuando se recibe un mensaje
    def _processMessage(self, message, idMessage, object):
        try:
            msg = self._compressor.decompress(message)
            self.__invokeOnMessageReceived(msg)
        except Exception as ex:
            loggerManager.get_endPoints_logger().exception("Error al procesar el mensaje")
            raise ex
        finally:
            if self._useAck:
                loggerManager.get_endPoints_logger().debug("Enviando ACK %s" % message)
                self._sendAck(object, idMessage)


    ## Metodo que se las clases hijas deben sobrescribir para realizar la recepcion
    # No se debe dejar bloqueado el hilo
    def _startReceive(self):
        pass

    ## Metodo al que se llama para inicializar la recepcion, sobrescribir en caso que sea necesario
    def _initializeReceive(self):
        pass

    ## Metodo al que se llama para liberar recursos de la recepcion, sobrescribir  en caso que sea necesario
    def _terminateReceive(self):
        pass

    ## Metodo para mandar el ack una vez que el mensaje se ha leido
    def _sendAck(self, object, idMessage):
        pass

    ## Metodo para detener la recepcion de mensajes
    def _stopReceive(self):
        pass


# author = sbermudel
# package description
from clitellum.core import compressors
from clitellum.endpoints.channels import reconnectiontimers
from clitellum.endpoints.channels.amqp.channels import OutBoundAmqpChannel, InBoundAmqpChannel
from clitellum.endpoints.channels.basechannels import Channel
from clitellum.endpoints.channels.tcpsocket import OutBoundChannelTcp, InBoundChannelTcp
from clitellum.endpoints.channels.zeromq import OutBoundChannelZeroMq, InBoundChannelZeroMq
from clitellum.endpoints.channels.amqp.builders import AmqpChannelBuilderFactory


## Crea un canal a partir de una configuracion
# { type : 0Mq,
#   timer : Logarithmic,
#   host : tcp://server:8080,
#   maxReconnections : 20,
#   compressor: { type="gzip", compressionLevel: 9 },
#   useAck : False}
def CreateOutBoundChannelFromConfig(config):
    if 'timer' in config:
        timer = reconnectiontimers.CreateTimerFormType(config["timer"])
    else:
        timer = reconnectiontimers.CreateConstantTimer()

    if 'compressor' in config:
        compressor = compressors.CreateCompressorFromConfig(config["compressor"])
    else:
        compressor = compressors.DefaultCompressor()

    maxReconnections = Channel.MAX_RECONNECTIONS
    if not config.get("maxReconnections") is None:
        maxReconnections = int(config.get("maxReconnections"))

    useAck = False
    if not config.get("useAck") is None:
        useAck = bool(config.get("useAck"))

    user = 'guest'
    password = 'guest'

    if 'user' in config:
        user = config['user']

    if 'password' in config:
        password = config['password']

    if config.get("type").lower() == "0mq":
        channel = OutBoundChannelZeroMq(config["host"], reconnectionTimer=timer,
                                        maxReconnections=maxReconnections, compressor=compressor)
    elif config.get("type").lower() == "tcp":
        channel = OutBoundChannelTcp(config["host"], reconnectionTimer=timer,
                                     maxReconnections=maxReconnections, compressor=compressor, useAck=useAck)
    elif config.get("type").lower() == "amqp":

        builder = AmqpChannelBuilderFactory.create_out_bound_builder()
        builder.set_uri(config["host"]) \
                .set_compressor(compressor) \
                .set_max_recconections(maxReconnections) \
                .set_reconnection_timer(timer) \
                .set_credentials(user, password) \
                .set_use_ack(useAck)

        if 'exchange_type' in config:
            if config['exchange_type'] == 'direct':
                builder.set_direct_exchange()
            else:
                builder.set_topic_exchange()
        else:
            builder.set_topic_exchange()

        if 'create_exchange' in config:
            builder.set_create_exchange(config['create_exchange'])
        else:
            builder.set_create_exchange(True)

        channel = builder.build()

    elif config.get("type").lower() == "custom":
        factory = _get_class(config["factory"])()
        channel = factory.create_from_cfg(config, timer, compressor, maxReconnections)

    else:
        channel = OutBoundChannelTcp(config["host"], reconnectionTimer=timer,
                                     maxReconnections=maxReconnections, compressor=compressor, useAck=useAck)
    return channel


def CreateInBoundChannelFromConfig(config):
    if 'timer' in config:
        timer = reconnectiontimers.CreateTimerFormType(config["timer"])
    else:
        timer = reconnectiontimers.CreateConstantTimer()

    if 'compressor' in config:
        compressor = compressors.CreateCompressorFromConfig(config["compressor"])
    else:
        compressor = compressors.DefaultCompressor()

    maxReconnections = Channel.MAX_RECONNECTIONS
    if not config.get("maxReconnections") is None:
        maxReconnections = int(config.get("maxReconnections"))

    useAck = False
    if 'useAck' in config:
        useAck = bool(config.get("useAck"))

    user = 'guest'
    password = 'guest'

    if 'user' in config:
        user = config['user']

    if 'password' in config:
        password = config['password']        

    maxChannelThread = 5
    if 'maxChannelThread' in config:
        maxChannelThread = config['maxChannelThread']

    if config.get("type").lower() == "0mq":
        channel = InBoundChannelZeroMq(config["host"], reconnectionTimer=timer,
                                       maxReconnections=maxReconnections, compressor=compressor)
    elif config.get("type").lower() == "tcp":
        channel = InBoundChannelTcp(config["host"], reconnectionTimer=timer,
                                    maxReconnections=maxReconnections, \
                                    compressor=compressor, useAck=useAck)
    elif config.get("type").lower() == "amqp":

        builder = AmqpChannelBuilderFactory.create_in_bound_builder()
        builder.set_uri(config["host"]) \
                .set_compressor(compressor) \
                .set_max_recconections(maxReconnections) \
                .set_reconnection_timer(timer) \
                .set_credentials(user, password) \
                .set_use_ack(useAck) \
                .set_max_threads(maxChannelThread) \

        if 'exchange_type' in config:
            if config['exchange_type'] == 'direct':
                builder.set_direct_exchange()
            else:
                builder.set_topic_exchange()
        else:
            builder.set_topic_exchange()
                
        if 'create_exchange' in config:
            builder.set_create_exchange(config['create_exchange'])
        else:
            builder.set_create_exchange(True)

        if 'create_queue' in config:
            builder.set_create_queue(config['create_queue'])
        else:
            builder.set_create_queue(True)

        channel = builder.build()

    elif config.get("type").lower() == "custom":
        factory = _get_class(config["factory"])()
        channel = factory.create_from_cfg(config, timer, compressor, maxReconnections)

    else:
        channel = InBoundChannelTcp(config["host"], reconnectionTimer=timer,
                                    maxReconnections=maxReconnections, \
                                    compressor=compressor, useAck=useAck)
    return channel


def _get_class(class_name):
    parts = class_name.split('.')
    module = ".".join(parts[:-1])
    module_obj = __import__(module)
    for comp in parts[1:]:
        module_obj = getattr(module_obj, comp)
    return module_obj


class CustomFactoryBase:
    def __init__(self):
        pass

    def create_from_cfg(self, cfg, reconnectionTimer, compressor, maxReconnections):
        pass



from clitellum.core import serialization
from clitellum.core.bus import MessageBuilder
from clitellum.endpoints import gateways
from clitellum.server import Identification

__author__ = 'Sergio'


def create_agent_from_config(cfg):
    identification = Identification(cfg['identification']['id'], cfg['identification']['type'])
    sg = gateways.CreateSenderFromConfig(cfg["sender_gateway"])
    return Publisher(identification, sg)


class Publisher:

    def __init__(self, identification, sender_gateway):
        self._identification = identification
        self._senderGateway = sender_gateway
        self._senderGateway.connect()

    @property
    def identification(self):
        return self._identification

    def publish(self, message, key, context=None):
        message_bus = MessageBuilder.create() \
                        .set_body(message) \
                        .set_type(key) \
                        .set_context(context) \
                        .set_service(self.identification) \
                        .build()

        message_serialized = serialization.dumps(message_bus)
        self._senderGateway.send(message_serialized, key)

    def __del__(self):
        self._senderGateway.close()



import logging

__author__ = 'sergio'


def getlogger():
    return logging.getLogger("clitellum")


def get_core_logger():
    return logging.getLogger("clitellum.core")


def get_endPoints_logger():
    return logging.getLogger("clitellum.endpoints")


def get_processors_logger():
    return logging.getLogger("clitellum.processors")



# author = sbermudel
# package description
from datetime import datetime
import os
import struct
import threading

from time import sleep, time
from pymongo import *

try:
    from thread import get_ident
except ImportError:
    from dummy_thread import get_ident


## Crea una cola desde un objeto de configuracion
#  { type : "MongoDb"
#     host: "mongodb://db1.example.net,db2.example.net:2500/?replicaSet=test&connectTimeoutMS=300000&maxPoolSize=20",
#     database: "MiBaseDeDatos",
#     collection: "MiColeccion" }
def CreateQueueFromConfig(config):
    if config.get("type").lower() == "mongodb":
        return MongoQueue(connection=config["host"], dbName=config['database'], collection=config['collection'])
    else:
        return MongoQueue(connection=config["host"], dbName=config['database'], collection=config['collection'])


def CreateMongoDbQueue(host, dbName, collection, auto_purge=True):
    return MongoQueue(host, dbName, collection, auto_purge=auto_purge)


## Clase que base que contiene los metodos basicos de cola
# con persistencia
class PersistenceQueue:
    def __init__(self, connection):
        self.__connection = connection

    @property
    def count(self):
        return 0

    @property
    def connectionString(self):
        return self.__connection

    def purge(self):
        pass

    def append(self, obj):
        pass

    def popleft(self, sleep_wait=True):
        pass

    def task_done(self):
        pass

    def task_not_done(self):
        pass

    def peek(self):
        pass


class MongoQueue(PersistenceQueue):
    def __init__(self, connection, dbName, collection, auto_purge=True):
        PersistenceQueue.__init__(self, connection)
        self._connectionString = connection
        self._colName = collection
        self._dbName = dbName

        self._conn = MongoClient(self._connectionString)
        if not self._colName in self._conn[self._dbName].collection_names():
            self._conn[self._dbName].create_collection(self._colName)

        self._collection = self._conn[self._dbName][self._colName]
        if auto_purge:
            self._collection.ensure_index( [("ProcessedAt", 1), ], name='_idx_Processed', expireAfterSeconds=3600)
            #else:
            #TODO: Eliminar el indice si no tiene auto_purge

        self._collection.ensure_index( [("Status", 1), ], name='_idx_Status')
        self._collection.update({"Status":1}, { "$set": {"Status":0, "ProcessedAt": None}}, multi=True)

        self._localStorage = threading.local()

    @property
    def count(self):
        return self._collection.count({"Status" : 0})

    @property
    def path(self):
        return self._connectionString

    def purge(self):
        self._collection.drop()

    def append(self, obj):
        self._collection.insert({ "Value" : obj, "Status" : 0, "ProcessedAt": None})

    def popleft(self, sleep_wait=True, timeout=-1):
        keep_pooling = True
        wait = 0.1
        max_wait = 2
        tries = 0

        start = time()

        while keep_pooling:
            rec = self._collection.find_and_modify(query={"Status" : 0}, update={ "$set": {"Status":1}})
            if not rec is None:
                self._localStorage.rec = rec['_id']
                return rec["Value"]
            else:
                if not sleep_wait:
                    keep_pooling = False
                    continue
                runningTime = time() - start
                if 0 < timeout < runningTime:
                    keep_pooling = False
                    continue
                tries += 1
                sleep(wait)
                wait = min(max_wait, tries/10 + wait)

        return None

    def task_done(self):
        self._collection.find_and_modify(query={"_id" : self._localStorage.rec},
                                    update={ "$set": {"Status":2, "ProcessedAt": datetime.utcnow()}})

    def task_not_done(self):
        self._collection.find_and_modify(query={"_id" : self._localStorage.rec},
                                    update={ "$set": {"Status":0, "ProcessedAt": None}})
    def peek(self):
        rec = self._collection.find_one(query={"Status" : 0})
        if not rec is None:
            return rec["Value"]
        else:
            return None

    def __del__(self):
        self._conn.close()

class FileQueue(PersistenceQueue):

    def __init__(self, connection):
        PersistenceQueue.__init__(self, connection)
        self.__connection = os.path.abspath(connection)
        if not os.path.exists(os.path.dirname(self.__connection)):
            os.makedirs(os.path.dirname(self.__connection))

        self.__fileWriter = open(self.__connection, "wb")
        self.__fileReader = open(self.__connection, "r+b")
        self.__fileAck = open(self.__connection, "r+b")

        self._localStorage = threading.local()
        self.__writerLock = threading.Lock()
        self.__readerLock = threading.Lock()
        self.__ackLock = threading.Lock()

    @property
    def count(self):
        return 0

    @property
    def connectionString(self):
        return self.__connection

    def purge(self):
        pass

    def append(self, obj):
        self.__writerLock.acquire()
        item = str(obj)
        buffer = struct.pack('!I%ds' % (len(item),), len(item), item)
        self.__fileWriter.write(buffer)
        self.__fileWriter.flush()
        self.__writerLock.release()

    def popleft(self, sleep_wait=True, timeout=-1):
        keep_pooling = True
        wait = 0.1
        max_wait = 2
        tries = 0

        start = time()

        while keep_pooling:
            rec = self.__readFromQueue()
            if not rec is None:
                self._localStorage.rec = rec['Id']
                return rec["Value"]
            else:
                if not sleep_wait:
                    keep_pooling = False
                    continue
                runningTime = time() - start
                if 0 < timeout < runningTime:
                    keep_pooling = False
                    continue
                tries += 1
                sleep(wait)
                wait = min(max_wait, tries/10 + wait)

        return None

    def __readFromQueue(self):
        self.__readerLock.acquire()
        item = None
        itemPos = self.__fileReader.tell()
        while True:
            buffer = self.__fileReader.read(4)
            if buffer != '':
                length = struct.unpack('>I', buffer)[0]
                item = self.__fileReader.read(length).strip('\0')

                if item != '':
                    break
            else:
                break

        self.__readerLock.release()
        return {"Id": itemPos, "Value": item}

    def task_done(self):
        self.__ackLock.acquire()
        self.__fileAck.seek(self._localStorage.rec)
        buffer = self.__fileAck.read(4)
        if buffer != '':
            length = struct.unpack('>I', buffer)[0]
            empty = struct.pack('!%ds'%length, "")
            self.__fileAck.write(empty)
            self.__fileAck.flush()

        self.__ackLock.release()

    def task_not_done(self):
        self.__readerLock.acquire()
        self.__fileReader.seek(self._localStorage.rec)
        self.__readerLock.release()

    def peek(self):
        pass

    def __del__(self):
        self.__fileReader.close()
        self.__fileWriter.close()
        self.__fileAck.close()



from clitellum.processors import builders
from clitellum.core.fsm import Startable
from clitellum.handlers import HandlerManager
from clitellum.core.bus import Identification

__author__ = 'sergio'


## Crea un server desde un config
def create_server_from_config(cfg):
    identification = Identification(cfg['identification']['id'], cfg['identification']['type'])
    ag = builders.create_agent_from_config(identification, cfg["agent"])
    control = builders.create_agent_from_config(identification, cfg["controller"])
    return Server(identification, ag, control)

class Server(Startable):

    def __init__(self, identification, agent, controller):
        Startable.__init__(self)
        self.identification = identification
        self.handler_manager = HandlerManager()
        self._agent = agent
        self._agent.configure(self.handler_manager)
        self._controller = controller

    def _invokeOnStart(self):
        Startable._invokeOnStart(self)
        self._agent.start()
        self._controller.start()

    def _invokeOnStopped(self):
        Startable._invokeOnStopped(self)
        self._agent.stop()
        self._controller.stop()

    def __del__(self):
        Startable.__del__(self)





import re
import json
import datetime
from dateutil import parser

__author__ = 'sbermudel'

DEFAULT_DATE_FORMAT = '%Y-%m-%dT%H:%M:%S.%f%z'
DEFAULT_ARGUMENT = "datetime_format"


class JsonClitellumEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        else:
            return super(JsonClitellumEncoder, self).default(obj)


def loads(s, **kwargs):
    format = kwargs.pop(DEFAULT_ARGUMENT, None) or DEFAULT_DATE_FORMAT
    source = json.loads(s, **kwargs)

    return iteritems(source, format)


def iteritems(source, format):
    for k, v in source.items():
        if isinstance(v, dict):
            iteritems(v, format)
        elif isinstance(v, str):
            match = re.search('^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[0-1]|0[1-9]|[1-2][0-9])?T(2[0-3]|[0-1][0-9]):([0-5][0-9]):([0-5][0-9])(\.[0-9]+)??(Z|[+-](?:2[0-3]|[0-1][0-9]):[0-5][0-9])?$', v)
            if not match is None:
                source[k] = parser.parse(v)
                pass
            else:
                try:
                    source[k] = datetime.datetime.strptime(v, format)
                except:
                    pass

    return source


def dumps(obj, skipkeys=False, ensure_ascii=True, check_circular=True,
          allow_nan=True, cls=JsonClitellumEncoder, indent=None, separators=None,
          encoding='utf-8', default=None, **kw):
    return json.dumps(obj, skipkeys=skipkeys, ensure_ascii=ensure_ascii, check_circular=check_circular,
        allow_nan=allow_nan, cls=cls, indent=indent, separators=separators,
        default=default, **kw)



import traceback

__author__ = 'sergio'


def create_error_message(message, exception):
    error_message = dict()
    error_message['message'] = message
    error_message['exception'] = dict()
    error_message['exception']['message'] = str(exception)
    error_message['exception']['traceback'] = traceback.format_exc()
    return error_message



__author__ = 'sbermudel'
## @package clitellum.endpoints.channels.events
#  Este paquete contiene las clases que definen los argumentos de los eventos


## Clase que contiene la informacion de un error de conexion
class ConnectionErrorArgs:
    ## Crea una instancia de ConnectionErrorArgs
    # @param host Nombre del host
    # @param numReconnections Numero de reconexiones
    # @param maxReconnections Numero maximo de reconexiones permitidas
    # @param timer Temporizador de reconexiones
    def __init__(self, channel, numReconnections= 0, maxReconnections = 0, timer = None):
        self.maxReconections = maxReconnections
        self.numReconnections = numReconnections
        self.channel = channel
        self.timer = timer

    ## Convierte la instancia a string
    def __str__(self):
        return "Host: %s\nReconnections: %s\n MaxReconnections: %s\n Timer %s " % \
               (self.channel.getHost(), self.numReconnections, self.maxReconections, self.timer)

    def __del__(self):
        del self.maxReconections
        del self.numReconnections


## Clase que contiene la informacion de un error en el envio de un mensaje
class SendErrorArgs:

    ## Crea una instancia de SendError
    def __init__(self, channel, message =""):
        self.channel = channel
        self.message = message

    ## Convierte la instancia a string
    def __str__(self):
        return "Host: %s\nMessage: %s" % \
                       (self.channel.getHost(), self.message)

    def __del__(self):
        del self.message


## Clase que contiene la informacion de un error en el envio de un mensaje
class SentMessageArgs:
    ## Crea una instancia de SendError
    def __init__(self, channel, message =""):
        self.channel = channel
        self.message = message

    ## Convierte la instancia a string
    def __str__(self):
        return "Host: %s\nMessage: %s" %\
               (self.channel.getHost(), self.message)

    def __del__(self):
        del self.message


## Clase que contiene la informacion de un error en el envio de un mensaje
class MessageReceivedArgs:

    ## Crea una instancia de SendError
    def __init__(self, message = None):
        self.message = message

    ## Convierte la instancia a string
    def __str__(self):
        return "Host: %s\nMessage: %s" % \
                       (self.message)

    def __del__(self):
        del self.message



import json
from clitellum.core.compressors import GzipCompressor

__author__ = 'lfernandm'


class MessageParser:
    def __init__(self):
        pass

    @classmethod
    def _byterize(cls, value, numBytes):
        displacements = range(0, numBytes * 8, 8)
        return [value >> i & 0xff for i in displacements]

    @classmethod
    def _unByterize(cls, byterizedValue):
        displacements = range(0, (len(byterizedValue)) * 8, 8)
        return sum(0x00000000000000000000000000000000 + byterizedValue[i] << displacements[i] for i in range(0, len(byterizedValue)))

    @classmethod
    def ToBytes(cls, message):
        serializedMessage = json.dumps(message)
        bytesMessage = GzipCompressor().compress(serializedMessage)
        bytesBodyType = message['Header']['BodyType'].encode('utf-8')
        callerContext = None
        callerContextL = len(message['Header']['CallStack']['_array'])
        if callerContextL > 0:
            callerContext = message['Header']['CallStack']['_array'][callerContextL - 1]
        idBytes = bytearray()
        typeBytes = bytearray()
        if callerContext is not None:
            idBytes = callerContext['Identification']['Id'].encode('utf-8')
            typeBytes = callerContext['Identification']['Type'].encode('utf-8')
        bMessage = bytearray()
        bMessage.append(message['Header']['Type'])
        bMessage.append(message['Header']['Priority'])
        bMessage.append(0)
        ticks = int(message['Header']['CreatedAt'])
        bMessage.extend(cls._byterize(ticks, 8))
        bMessage.extend(cls._byterize(len(bytesBodyType), 4))
        bMessage.extend(bytesBodyType)
        bMessage.extend(cls._byterize(len(bytesMessage), 4))
        bMessage.extend(bytesMessage)
        bMessage.extend(cls._byterize(len(idBytes), 4))
        if callerContext is not None:
            bMessage.extend(idBytes)
            bMessage.extend(cls._byterize(len(typeBytes), 4))
            bMessage.extend(typeBytes)
        return bMessage

    @classmethod
    def GetMessageFromBytes(cls, binMessage):
        compressor = GzipCompressor()
        bodyTypeLength = cls._unByterize(binMessage[11:15])
        messagePosition = 15 + bodyTypeLength
        messageLength = cls._unByterize(binMessage[messagePosition:messagePosition + 4])
        return compressor.decompress(binMessage[messagePosition + 4:messagePosition + 4 + messageLength])



__author__ = 'sbermudel'
## @package clitellum.endpoints.channels.exceptions
#  Este paquete contiene las clases de error de los channels


## Excepion que se produce cuando hay un error en la conexion
class ConnectionError(Exception):

    ## Crea una instancia de ConnectionError
    # @param message Descripcion del error
    def __init__(self, message):
        self.message = message

    ## Conviente la instancia a string
    def __str__(self):
        return repr(self.message)

## Excepion que se produce cuando hay un error en el envio de mensajes
class SendError(Exception):

    ## Crea una instancia de SendError
    # @param message Descripcion del error
    def __init__(self, message):
        self.message = message

    ## Conviente la instancia a string
    def __str__(self):
        return repr(self.message)



from setuptools import setup

setup(
    name='clitellum',
    version='4.0.4',
    packages=['clitellum', 'clitellum.core', 'clitellum.endpoints', 'clitellum.processors',
              'clitellum.endpoints.channels', 'clitellum.endpoints.channels.amqp'],
    package_dir={'clitellum': 'clitellum'},
    url='https://gitlab.com/clitellum/clitellum/wikis/home',
    license='GPL',
    author='Sergio.Bermudez',
    author_email='sbermudezlozano@gmail.com',
    description='Clitellum Microservice Framework',
    extras_require={
    },
    install_requires=['pika', 'pymongo','zmq', 'pyyaml', 'fluidity-sm', 'python-dateutil']
)



# coding=utf-8
# author = sbermudel
## @package clitellum.core.loadbalancers
# Este paquete contiene los balanceadores de carga
import itertools
import threading


## Devuelve el balanceador de carga por defecto
def __defaultLoadBalancer():
    return RoundRobinLoadBalancer()


## Crea un balanceador de carga desde una configuracion
# { type: "RoundRobin" }
def CreateRouterFromConfig(config):
    if config is None:
        return __defaultLoadBalancer()

    if config.get("type") == "RoundRobin":
        return RoundRobinLoadBalancer()
    else:
        return __defaultLoadBalancer()


## Clase que contiene la informacion necesaria para balancear la carga de un canal
class LoadBalancerItem:
    def __init__(self, item, balancingInfo):
        self.item = item
        self.balancingInfo = balancingInfo


## Clase base que define el comportamiento de un balanceador de carga.
# El balanceador de carga es el encargado de distribuir el mensaje en los diferentes canales.
# El tipo de balanceamiento puede ser "Round Robin", etc.
# Al añadir un canal al balanceador, se le indica la informacion de balanceo, especifica para cada
# motor de balanceo.
class LoadBalancer:
    ## Crea una instancia de Router
    def __init__(self):
        self._items = list()
        self._initialized = threading.Event()
        self._load()

    ## Añade un canal con la informacion especifica en al router.
    # @param channel Instancia de OutBoundChannel
    # @param routingInfo Informacion de enrutamiento especifica de cada tipo de motor
    # @param resetAfter Indica si despues de añadir el canal se debe hacer un reset del router
    def addChannel(self, channel, balancingInfo=None, resetAfter=True):
        item = LoadBalancerItem(channel, balancingInfo)
        self._items.append(item)
        if resetAfter:
            self.reset()

    ## Devuelve el siguiente canal al que se le debe enviar el mensaje
    def next(self):
        self._initialized.wait()
        return self._next()

    ## Reinicia el router
    def reset(self):
        self._initialized.clear()
        self._load()
        self._initialized.set()

    ## Metodo que debe ser sobre escrito por cada motor de enrutamiento para reiniciar el router
    def _load(self):
        pass

    ## Metodo que debe ser sobre escrito por cada motor de enrutamiento para enrutar el mensaje
    def _next(self):
        pass


## Clase que implementa un balanceador de carga de tipo round robin
class RoundRobinLoadBalancer(LoadBalancer):

    ## Crea una instancia de RoundRobinLoadBalancer
    def __init__(self):
        LoadBalancer.__init__(self)

    def _load(self):
        self._ciclo = itertools.cycle(self._items)

    def _next(self):
        return next(self._ciclo).item




from clitellum.core import serialization

__author__ = 'Sergio'


class HandlerManager:

    def __init__(self):
        self.__handlers = dict()

    def handler(self, key, **options):
        def decorator(f):
            self.add_handler(key, f, options)
            return f
        return decorator

    def add_handler(self, key, class_type, options):
        self.__handlers[key] = class_type

    def get_handler(self, key):
        return self.__handlers[key]()


class HandlerBase:

    def __init__(self):
        self.__bus = None
        self.__header = None
        self.__context = None

    def initialize(self, bus, header):
        self.__bus = bus
        self.__header = header
        self.__context = header['CallContext']

    @property
    def bus(self):
        """
        :rtype: SenderGateway
        :return:
        """
        return self.__bus

    @property
    def context(self):
        return self.__context

    def handle_message(self, message):
        pass

    def _send(self, message, key):
        self.bus.send(message, key=key, header=self.__header)

    def _reply(self, message, key):
        self.bus.reply(message, key, header=self.__header)



# author = sbermudel
# package description
import time


## Temoraizador de espera hasta que se cumple la condicion del predicado
def waitUntil(somepredicate, timeout, *args):
    period = 0.25
    mustEnd = time.time() + timeout
    while time.time() < mustEnd:
        if somepredicate(args): return True
        time.sleep(period)
    return False
